@@ -1,52 +1,51 @@
 """Iterative datapipes toread weather dataset in seqrecord format"""
 
-from typing import Callable, Dict, List, Optional, Tuple
+from typing import Callable, Dict, Iterator, List, Optional, Tuple
 
 import numpy as np
 import torch
-import torchdata.datapipes as dp
+import torchdata.datapipes as datapipe
+from torchdata.datapipes.iter import IterableWrapper, IterDataPipe
 from tqdm import tqdm
 
 from .constants import VAR_LEVEL_TO_NAME_LEVEL
 from .seqrecord import WSeqRecord
 
 
-@dp.functional_datapipe("weather_datapipe")
-class WeatherDatapipeFromWSeqRecord(dp.iter.IterDataPipe):
+@datapipe.functional_datapipe("gen_framepair")
+class FramePairsFromWSeqRecord(datapipe.iter.IterDataPipe):
     """A torch datapiple class that iteratively read frame pairs from weather dataset (encoded by WSeqRecord)."""
 
     def __init__(
         self,
-        record: WSeqRecord,
-        input_features: List[str],
-        target_features: List[str],
-        max_pred_steps: int,
+        source_dp: datapipe.iter.IterDataPipe,
     ) -> None:
         super().__init__()
-        self.record = record
-        self.input_features = input_features
-        self.target_features = target_features
-        self.max_pred_steps = max_pred_steps
+        self.source_dp = source_dp
 
     def __iter__(self):
-        # naive iterate
-        # yield from self.record.iterate_frame_pairs(
-        #     self.input_features, self.target_features, self.max_pred_steps
-        # )
-        # prefetech
-        # yield from self.record.fetch_frame_pairs(
-        #     self.input_features,
-        #     self.target_features,
-        #     self.max_pred_steps,
-        #     prefetch_buffer_size=10,
-        # )
-        yield from self.record.async_iterate_frame_pairs(
-            self.input_features,
-            self.target_features,
-            self.max_pred_steps,
+        yield from WSeqRecord.iterate_framepairs_from_files(
+            self.source_dp,
         )
 
 
+@datapipe.functional_datapipe("gen_fileidx")
+class FileidxFromWSeqRecord(datapipe.iter.IterDataPipe):
+    """A torch datapiple class that iteratively read fileidx from weather dataset (encoded by WSeqRecord)."""
+
+    def __init__(
+        self,
+        record_dp: datapipe.iter.IterDataPipe,
+    ) -> None:
+        super().__init__()
+        self.record_dp = record_dp
+
+    def __iter__(self):
+        for record in self.record_dp:
+            for fileidx in range(record.num_files):
+                yield record, fileidx
+
+
 def collate_fn(batch: List[Dict[str, np.ndarray]]) -> Dict[str, torch.Tensor]:
     collated_batch: Dict[str, torch.Tensor] = {}
     collated_batch["input"] = torch.from_numpy(
@@ -67,35 +66,40 @@ def collate_fn(batch: List[Dict[str, np.ndarray]]) -> Dict[str, torch.Tensor]:
     return collated_batch
 
 
-def build_datapipes(
-    datapipe: dp.iter.IterDataPipe,
-    shuffle_buffer_size: Optional[int],
+def build_wdatapipe(
+    records: List[WSeqRecord],
+    file_shuffle_buffer_size: Optional[int],
+    data_shuffle_buffer_size: Optional[int],
     batch_size: int,
     mappings: List[Callable],
-) -> dp.iter.IterDataPipe:
+) -> datapipe.iter.IterDataPipe:
     """Iteratively apply operations to datapipe: shuffle, sharding, map, batch, collator
 
     Args:
-        datapipe (dp.datapipe.IterDataPipe): entry datapipe
+        datapipe (datapipe.datapipe.IterDataPipe): entry datapipe
         shuffle_buffer_size (Optional[int]): buffer size for pseudo-shuffle
         batch_size (int):
         mappings (List[Callable]): a list of transforms applied to datapipe, between sharding and batch
 
     Returns:
-        dp.datapipe.IterDataPipe: transformed datapipe ready to be sent to dataloader
+        datapipe.datapipe.IterDataPipe: transformed datapipe ready to be sent to dataloader
     """
+    dp = IterableWrapper(records).gen_fileidx()
     # Shuffle will happen as long as you do NOT set `shuffle=False` later in the DataLoader
     # https://pytorch.org/data/main/tutorial.html#working-with-dataloader
-    if shuffle_buffer_size is not None:
-        datapipe = datapipe.shuffle(buffer_size=shuffle_buffer_size)
+    if file_shuffle_buffer_size is not None:
+        dp = dp.shuffle(buffer_size=file_shuffle_buffer_size)
     # sharding: Place ShardingFilter (datapipe.sharding_filter) as early as possible in the pipeline,
     # especially before expensive operations such as decoding, in order to avoid repeating these expensive operations across worker/distributed processes.
-    datapipe = datapipe.sharding_filter()
+    # output will be a sequence of file_idx(s) distributed to different workers (
+    dp = dp.sharding_filter()
+    dp = dp.gen_framepair()
+    dp = dp.shuffle(buffer_size=data_shuffle_buffer_size)
     for i, mapping in enumerate(mappings):
-        datapipe = datapipe.map(fn=mapping)
+        dp = dp.map(fn=mapping)
     # Note that if you choose to use Batcher while setting batch_size > 1 for DataLoader,
     # your samples will be batched more than once. You should choose one or the other.
     # https://pytorch.org/data/main/tutorial.html#working-with-dataloader
-    datapipe = datapipe.batch(batch_size=batch_size, drop_last=True)
-    datapipe = datapipe.collate(collate_fn=collate_fn)
-    return datapipe
+    dp = dp.batch(batch_size=batch_size, drop_last=True)
+    dp = dp.collate(collate_fn=collate_fn)
+    return dp
