@@ -22,11 +22,14 @@ from typing import (
     Tuple,
     TypeVar,
     Union,
+    Iterator,
 )
 
 import aiofiles
 import numpy as np
 import yaml
+import subprocess
+import shutil
 
 from seqrecord.utils import (
     CONSUMER_SLEEP_INTERVAL,
@@ -37,22 +40,14 @@ from seqrecord.utils import (
     WriterBuffer,
 )
 
-MAX_RECORDFILE_SIZE = 1e9  # 1e8, 100 mb, maximum size of a single record file
-
+MAX_RECORDFILE_SIZE = 2e9  # 1e9  # 1e8, 100 mb, maximum size of a single record file
+FILE_CACHE_SIZE = 10  # 10, maximum number of record files to keep in local disk
 WSR = TypeVar("WSR", bound="WSeqRecord")
 
-# todo: record file size and write buffer size, how to choose it, and tune it
-# todo: test effect of caching
-
-
-# todo: use azure uri and manage sources of storage account by ourselves (replica of dataset)
+# todo: add property stuff for attributes not to be changed by user
 # todo: MISSING, some transform work, subsample etc. Need to make sure data is same as produced by existing dataset
 
 
-def recordfileidx2path(recorddir: str, file_idx: int) -> str:
-    return os.path.join(recorddir, f"record_{file_idx}.bin")
-
-
 class _PrefetchData:
     def __init__(self, source_data_generator, buffer_size: int):
         self.run_prefetcher = True
@@ -66,10 +61,7 @@ class _PrefetchData:
 class WSeqRecord:
     """A serialization protocal that stores a single continuous long sequence of weather data into record files, while provides metadata of each frame to enable efficient random access of frames."""
 
-    def __init__(
-        self,
-        recorddir: str,
-    ) -> None:
+    def __init__(self, recorddir: str, local_cache_dir: Optional[str] = None) -> None:
         """_summary_
 
         Args:
@@ -78,112 +70,67 @@ class WSeqRecord:
         # folder that stores data in a separate directory (subfolder)
         self.recorddir: str = recorddir
         os.makedirs(self.recorddir, exist_ok=True)
+        # in case we use realtive path '~'
+        self.local_cache_dir = local_cache_dir
+        if local_cache_dir is not None:
+            self.local_cache_dir = os.path.abspath(os.path.expanduser(local_cache_dir))
+            if (
+                os.path.exists(self.local_cache_dir)
+                and len(os.listdir(self.local_cache_dir)) > 0
+            ):
+                print("Warning: local cache dir is not empty. Clearing it now.")
+                shutil.rmtree(self.local_cache_dir, ignore_errors=True)
+            os.makedirs(self.local_cache_dir, exist_ok=True)
         self.features_written = None
-        self.num_bytes: int = 0  # number of bytes in current file
-        # track the frame idx endpoints for each record file, [[start_idx, end_idx)], includes start, exclusive on end
-        self.idx_range_of_files: List[Tuple[int, int]] = []
-        # file object for current record file
-        self.file_desc: Optional[BinaryIO] = None
-        # serialization proto info of each data item
-        self.metadata: Dict[int, dict] = {}
-        # index of current data item to be processed
-        self.frame_idx: int = 0
 
-        # write buffer # todo: determine the size of write buffer
-        self.write_buffer = WriterBuffer()
-        self.timer = TimeTracker()
+    @staticmethod
+    def subfolder_name(rank: int, world_size: int) -> str:
+        return f"{rank}"
 
-    # todo add append_item(), add new frames to existing dataset (probably saved already)
-    # todo add preput worker
-    def write_item(
-        self,
-        frame: Dict[str, np.ndarray],
-    ) -> None:
-        """write one frame data dict(feature->np.ndarray) into byte repr and write bytes into
-        current file. It has to be called by users in a sequential order. Every key,value in frame will
-        be written into file.
+    @staticmethod
+    def fileidx2name(file_idx: int) -> str:
+        return f"record_{file_idx}.bin"
 
-        #todo: add default settings, where features are all features in metadata[idx]
+    def fileidx2path(self, file_idx: int, local_cache_dir: Optional[str] = None):
+        """Turn absolute file idx into relative path of the corresponding record file.
+        Write to self.local_dir and move.
 
         Args:
-            frame (Dict[str, np.ndarray]): feature to data (np.ndarray)
+            file_idx (int): _description_
+            local_cache_dir (str, optional): The directory to cache the record file. Defaults to None.
+        Returns:
+            _type_: _description_
+
+        Yields:
+            _type_: _description_
         """
-        if self.features_written is None:
-            self.features_written = [key for key in frame]
-        file_idx = len(self.idx_range_of_files) - 1
-        if self.num_bytes == 0:
-            # new file
-            self.idx_range_of_files.append([self.frame_idx])
-            file_idx += 1
-            self.file_desc = open(
-                recordfileidx2path(self.recorddir, file_idx),
-                mode="wb",
+        dir = self.recorddir if local_cache_dir is None else local_cache_dir
+        rank_id = self.meta_file[file_idx].get("rank_id", -1)
+        if rank_id == -1:
+            # there is no rank hierarachy
+            return os.path.join(dir, self.fileidx2name(file_idx))
+        else:
+            return os.path.join(
+                dir,
+                self.subfolder_name(rank_id, self.num_ranks),
+                f"record_{self.meta_file[file_idx]['rel_file_idx']}.bin",
             )
-            # todo: check here if we are appending, if it is, we need to open a new file
 
-        # get file name and start position for frame
-        self.metadata[self.frame_idx] = {
-            "frame_idx": self.frame_idx,
-            "file_idx": file_idx,
-            "bytes_offset": self.num_bytes,
-        }
-        num_bytes_in_frame = 0
-        for feature, data in frame.items():
-            self.metadata[self.frame_idx][feature] = {
-                "is_none": (
-                    data.dtype == np.dtype("O") and data == None
-                ),  # this feature is essentially missing, and
-                "dtype": data.dtype,
-                "shape": data.shape,
-                "bytes_offset": num_bytes_in_frame,
-                "nbytes": data.nbytes,
-            }
-            self.write_buffer.write(data.tobytes())
-            num_bytes_in_frame += data.nbytes
+    def recordfile_generator(self, frame_generator: Iterator):
+        """Ignore the complexity of rank/world size from multi-processing. This method only
+        focus on file/frame.
 
-        self.metadata[self.frame_idx]["nbytes"] = num_bytes_in_frame
-        self.num_bytes += num_bytes_in_frame
-        self.frame_idx += 1
+        Args:
+            frame_generator (callable): _description_
 
-        # if self.write_buffer.is_full():
-        #     self.file_desc.write(self.write_buffer.getbuffer())
-        #     self.write_buffer.clear()
-        #     self.file_desc.flush()
-
-        if self.num_bytes > MAX_RECORDFILE_SIZE:
-            # current file is big enough
-            self.num_bytes = 0
-            self.idx_range_of_files[-1].append(self.frame_idx)
-            self.file_desc.write(self.write_buffer.getbuffer())
-            self.write_buffer.clear()
-            self.file_desc.flush()
-            self.file_desc.close()
-        return
-
-    def close_recordfile(self):
-        """Close opened file descriptor!
-
-        This needs to be called by user when finishes scanning over the dataset.
+        Yields:
+            _type_: _description_
         """
-        if len(self.idx_range_of_files[-1]) == 1:
-            # the last record file has not reached maximum file size
-            self.idx_range_of_files[-1].append(self.frame_idx)
-
-            # write what's left in self.writer buffer
-            self.file_desc.write(self.write_buffer.getbuffer())
-            self.write_buffer.close()
-
-        self.file_desc.close()
-        self.write_buffer = None
-        self.file_desc = None
-        self.num_frames = self.frame_idx
-        self.num_files = len(self.idx_range_of_files)
-
-    def recordfile_generator(self, frame_generator: callable):
         try:
             write_buffer = WriterBuffer()
             num_bytes = 0
-            self.idx_range_of_files = []
+            self.meta_file = {}
+            self.meta_frame = {}
             frame_idx = 0
             file_idx = 0
             for frame in frame_generator:
@@ -191,15 +138,18 @@ class WSeqRecord:
                     self.features_written = [key for key in frame]
                 if num_bytes == 0:
                     # new file
-                    self.idx_range_of_files.append([frame_idx])
-                self.metadata[frame_idx] = {
-                    "frame_idx": frame_idx,
+                    self.meta_file[file_idx] = {
+                        "frame_idx_start": frame_idx,
+                        "relative_path": self.fileidx2name(file_idx),
+                    }
+                    # relative path to the record file does not contain directory of the corresponding seqrecord
+                self.meta_frame[frame_idx] = {
                     "file_idx": file_idx,
                     "bytes_offset": num_bytes,
                 }
                 num_bytes_in_frame = 0
                 for feature, data in frame.items():
-                    self.metadata[frame_idx][feature] = {
+                    self.meta_frame[frame_idx][feature] = {
                         "is_none": (
                             data.dtype == np.dtype("O") and data == None
                         ),  # this feature is essentially missing, and
@@ -211,24 +161,34 @@ class WSeqRecord:
                     write_buffer.write(data.tobytes())
                     num_bytes_in_frame += data.nbytes
 
-                self.metadata[frame_idx]["nbytes"] = num_bytes_in_frame
+                self.meta_frame[frame_idx]["nbytes"] = num_bytes_in_frame
                 frame_idx += 1
                 num_bytes += num_bytes_in_frame
                 if num_bytes > MAX_RECORDFILE_SIZE:
                     # current file is big enough
                     num_bytes = 0
-                    self.idx_range_of_files[-1].append(frame_idx)
+                    self.meta_file[file_idx]["frame_idx_end"] = frame_idx
                     write_buffer.clear()
-                    yield (file_idx, write_buffer.getvalue())
+                    yield (
+                        self.fileidx2path(file_idx),
+                        write_buffer.getvalue(),
+                    )
                     file_idx += 1
 
-            if len(self.idx_range_of_files[-1]) == 1:
+            if (
+                file_idx in self.meta_file
+                and self.meta_file[file_idx].get("frame_idx_end", None) is None
+            ):
                 # there is content left in the write_buffer
-                self.idx_range_of_files[-1].append(frame_idx)
-                yield (file_idx, write_buffer.getvalue())
+                self.meta_file[file_idx]["frame_idx_end"] = frame_idx
+                yield (
+                    self.fileidx2path(file_idx, self.local_cache_dir),
+                    write_buffer.getvalue(),
+                )
+                file_idx += 1
         finally:
             write_buffer.close()
-            self.num_files = len(self.idx_range_of_files)
+            self.num_files = file_idx
             self.num_frames = frame_idx
 
     def put_frame(self, frame_generator: callable, prefetch_buffer_size: int = 5):
@@ -246,19 +206,57 @@ class WSeqRecord:
                 daemon=True,
             )
             thread.start()
+            file_cache = []
+            subprocesses = deque()
             while prefetch_data.run_prefetcher:
                 if len(prefetch_data.prefetch_buffer) > 0:
-                    file_idx, content = prefetch_data.prefetch_buffer.popleft()
-                    with open(recordfileidx2path(self.recorddir, file_idx), "wb") as f:
+                    (
+                        file_path,
+                        content,
+                    ) = prefetch_data.prefetch_buffer.popleft()
+                    with open(file_path, "wb") as f:
                         f.write(content)
+                    file_cache.append(file_path)
+                    if (
+                        self.local_cache_dir is not None
+                        and len(file_cache) > FILE_CACHE_SIZE
+                    ):
+                        # move record files to recorddir in the background
+                        subprocesses.append(
+                            subprocess.Popen(
+                                [
+                                    "mv",
+                                ]
+                                + file_cache
+                                + [f"{self.recorddir}/"]
+                            )
+                        )
+                        file_cache = []
                 else:
                     # TODO: Calculate sleep interval based on previous availability speed
                     time.sleep(CONSUMER_SLEEP_INTERVAL)
+                    # todo: check if poll() is used properly
+                    if len(subprocesses) > 0 and subprocesses[0].poll() is not None:
+                        subprocesses.popleft()
+
         finally:
             prefetch_data.run_prefetcher = False
             if thread is not None:
                 thread.join()
                 thread = None
+            if self.local_cache_dir is not None and len(file_cache) > 0:
+                subprocesses.append(
+                    subprocess.Popen(
+                        [
+                            "mv",
+                        ]
+                        + file_cache
+                        + [f"{self.recorddir}/"]
+                    )
+                )
+                file_cache = []
+            for p in subprocesses:
+                p.wait()
 
     def read_frame(
         self,
@@ -302,24 +300,22 @@ class WSeqRecord:
         Yields:
             Generator[Dict[str, np.ndarray], None, None]: generates one-frame data
         """
-        assert all(
-            len(endpoints) == 2 for endpoints in self.idx_range_of_files
-        ), "Record file endpoints are not complete!"
-
         for file_idx in range(self.num_files):
             file_desc = open(
-                recordfileidx2path(self.recorddir, file_idx=file_idx), mode="rb"
+                os.path.join(self.recorddir, self.meta_file[file_idx]["relative_path"]),
+                mode="rb",
             )
-            endpoints = self.idx_range_of_files[file_idx]
-            for idx in range(endpoints[0], endpoints[1]):
-                frame = self.read_frame(file_desc, self.metadata[idx], features)
+            for idx in range(
+                self.meta_file[file_idx]["frame_idx_start"],
+                self.meta_file[file_idx]["frame_idx_end"],
+            ):
+                frame = self.read_frame(file_desc, self.meta_frame[idx], features)
 
                 yield {feature: frame[feature] for feature in features}
             file_desc.close()
 
-    # todoL add prefetch using threads or asyncio
     # todo: test effect of caching on real data
-    def iterate_frame_pairs(
+    def iterate_framepairs(
         self,
         input_features: List[str],
         target_features: List[str],
@@ -338,40 +334,50 @@ class WSeqRecord:
             Generator[Dict[str, np.ndarray], None, None]: data item [feature->data]. All data items are being returned sequentially
         """
         file_manager = FileManager(
-            fileidx2path=partial(recordfileidx2path, recorddir=self.recorddir),
             cache_capacity=filedesc_cache_cap,
         )
         # given that, input and target features do not overlap, we only cache target frame
         # LRU might not be suitable, evicting based on idx seems better
         frame_cache = LRUCache(frame_cache_cap)
-        assert all(
-            len(endpoints) == 2 for endpoints in self.idx_range_of_files
-        ), "Record file endpoints are not complete!"
-        num_frames = len(self.metadata)  # or self.num_frames
+
         for fileidx4input in range(self.num_files):
-            filedesc4input = file_manager.open_file(file_idx=fileidx4input)
-            endpoints = self.idx_range_of_files[fileidx4input]
+            filedesc4input = file_manager.open_file(
+                file_idx=fileidx4input,
+                file_path=os.path.join(
+                    self.recorddir, self.meta_file[fileidx4input]["relative_path"]
+                ),
+            )
+            endpoints = (
+                self.meta_file[fileidx4input]["frame_idx_start"],
+                self.meta_file[fileidx4input]["frame_idx_end"],
+            )
             # no target frame to predict for the last frame
             for frameidx4input in range(
                 endpoints[0],
-                min(endpoints[1], num_frames - 1),  # self.num_frames
+                min(endpoints[1], self.num_frames - 1),  # self.num_frames
             ):
                 input_frame = self.read_frame(
-                    filedesc4input, self.metadata[frameidx4input], input_features
+                    filedesc4input, self.meta_frame[frameidx4input], input_features
                 )
                 # get the target frame for prediction, both start, stop inclusive
                 lookahead_steps = min(
                     random.randint(1, max_pred_steps),
-                    num_frames - 1 - frameidx4input,
+                    self.num_frames - 1 - frameidx4input,
                 )
                 frameidx4target = frameidx4input + lookahead_steps
                 target_frame = frame_cache.get(frameidx4target)
                 if target_frame is None:
-                    fileidx4target = self.metadata[frameidx4target]["file_idx"]
-                    filedesc4target = file_manager.open_file(fileidx4target)
+                    fileidx4target = self.meta_frame[frameidx4target]["file_idx"]
+                    filedesc4target = file_manager.open_file(
+                        file_idx=fileidx4target,
+                        file_path=os.path.join(
+                            self.recorddir,
+                            self.meta_file[fileidx4target]["relative_path"],
+                        ),
+                    )
                     target_frame = self.read_frame(
                         filedesc4target,
-                        self.metadata[frameidx4target],
+                        self.meta_frame[frameidx4target],
                         target_features,
                     )
                 # colllate input and target frames so that input and target frame are np.ndarray
@@ -381,7 +387,6 @@ class WSeqRecord:
                 target_frame = np.vstack(
                     [target_frame[feature] for feature in target_features]
                 )
-                # print(self.timer.summarize())
                 yield {
                     "input": input_frame,
                     "target": target_frame,
@@ -391,7 +396,7 @@ class WSeqRecord:
                 }
         file_manager.close_all_files()
 
-    def async_iterate_frame_pairs(
+    def async_iterate_framepairs(
         self,
         input_features: List[str],
         target_features: List[str],
@@ -417,37 +422,48 @@ class WSeqRecord:
         file_desc_cache = LRUCache(capacity=filedesc_cache_cap)
         # given that, input and target features do not overlap, we only cache target frame
         # LRU might not be suitable, evicting based on idx seems better
-        assert all(
-            len(endpoints) == 2 for endpoints in self.idx_range_of_files
-        ), "Record file endpoints are not complete!"
 
-        num_frames = len(self.metadata)
         # read two frames using asyn io
         # file cache should only be used for future frames, since base file desc is used continuously
         try:
             for fileidx4input in range(self.num_files):
                 filedesc4input = None
-                endpoints = self.idx_range_of_files[fileidx4input]
+                endpoints = (
+                    self.meta_file[fileidx4input]["frame_idx_start"],
+                    self.meta_file[fileidx4input]["frame_idx_end"],
+                )
 
                 # no target frame to predict for the last frame
                 for frameidx4input in range(
-                    endpoints[0], min(endpoints[1], num_frames - 1)
+                    endpoints[0], min(endpoints[1], self.num_frames - 1)
                 ):
                     lookahead_steps = min(
                         random.randint(1, max_pred_steps),
-                        num_frames - 1 - frameidx4input,
+                        self.num_frames - 1 - frameidx4input,
                     )
                     frameidx4target = frameidx4input + lookahead_steps
-                    fileidx4target = self.metadata[frameidx4target]["file_idx"]
+                    fileidx4target = self.meta_frame[frameidx4target]["file_idx"]
 
                     filedesc4target = file_desc_cache.get(fileidx4target)
                     if filedesc4input is None and filedesc4target is None:
                         # both files need to be opened
                         file_descs = loop.run_until_complete(
                             asyncio.gather(
-                                async_open_file(fileidx4input, None, self.recorddir),
                                 async_open_file(
-                                    fileidx4target, file_desc_cache, self.recorddir
+                                    fileidx4input,
+                                    None,
+                                    os.path.join(
+                                        self.recorddir,
+                                        self.meta_file[fileidx4input]["relative_path"],
+                                    ),
+                                ),
+                                async_open_file(
+                                    fileidx4target,
+                                    file_desc_cache,
+                                    os.path.join(
+                                        self.recorddir,
+                                        self.meta_file[fileidx4target]["relative_path"],
+                                    ),
                                 ),
                             )
                         )
@@ -456,39 +472,44 @@ class WSeqRecord:
                     elif filedesc4input is None:
                         # only need files for input frame
                         file_descs = loop.run_until_complete(
-                            async_open_file(fileidx4input, None, self.recorddir)
+                            async_open_file(
+                                fileidx4input,
+                                None,
+                                os.path.join(
+                                    self.recorddir,
+                                    self.meta_file[fileidx4input]["relative_path"],
+                                ),
+                            )
                         )
                         filedesc4input = file_descs
                     elif filedesc4target is None:
                         # only need files for target frame
                         file_descs = loop.run_until_complete(
                             async_open_file(
-                                fileidx4target, file_desc_cache, self.recorddir
+                                fileidx4target,
+                                file_desc_cache,
+                                os.path.join(
+                                    self.recorddir,
+                                    self.meta_file[fileidx4target]["relative_path"],
+                                ),
                             )
                         )
                         filedesc4target = file_descs
 
-                    start_time = perf_counter()
                     frame_pairs = loop.run_until_complete(
                         asyncio.gather(
                             async_read_frame(
                                 filedesc4input,
-                                self.metadata[frameidx4input],
+                                self.meta_frame[frameidx4input],
                                 input_features,
                             ),
                             async_read_frame(
                                 filedesc4target,
-                                self.metadata[frameidx4target],
+                                self.meta_frame[frameidx4target],
                                 target_features,
                             ),
                         )
                     )
-                    end_time = perf_counter()
-                    self.timer.add(
-                        end_time - start_time,
-                        frame_pairs[0].nbytes + frame_pairs[1].nbytes,
-                    )
-                    # print(self.timer.summarize())
                     yield {
                         "input": frame_pairs[0],
                         "target": frame_pairs[1],
@@ -537,7 +558,7 @@ class WSeqRecord:
                 # TODO: Calculate sleep interval based on previous consumption speed
                 time.sleep(PRODUCER_SLEEP_INTERVAL)
 
-    def fetch_frame_pairs(
+    def fetch_framepairs(
         self,
         input_features: List[str],
         target_features: List[str],
@@ -546,7 +567,7 @@ class WSeqRecord:
     ):
 
         if prefetch_buffer_size < 1:
-            yield from self.iterate_frame_pairs(
+            yield from self.iterate_framepairs(
                 input_features, target_features, max_pred_steps
             )
         else:
@@ -554,7 +575,7 @@ class WSeqRecord:
             # preftech using a separate thread
             try:
                 prefetch_data = _PrefetchData(
-                    self.iterate_frame_pairs(
+                    self.iterate_framepairs(
                         input_features, target_features, max_pred_steps
                     ),
                     prefetch_buffer_size,
@@ -577,32 +598,106 @@ class WSeqRecord:
                     thread.join()
                     thread = None
 
-    def dump(self) -> None:
+    @staticmethod
+    def iterate_framepairs_from_files(
+        fileidx_generator: Iterator[int],
+        filedesc_cache_cap: int = 10,
+        frame_cache_cap: int = 20,
+    ) -> Generator[Dict[str, np.ndarray], None, None]:
+        file_manager = FileManager(
+            cache_capacity=filedesc_cache_cap,
+        )
+        frame_cache = LRUCache(frame_cache_cap)
+
+        for record, fileidx4input in fileidx_generator:
+            filedesc4input = file_manager.open_file(
+                file_idx=fileidx4input,
+                file_path=os.path.join(
+                    record.recorddir, record.meta_file[fileidx4input]["relative_path"]
+                ),
+            )
+            endpoints = (
+                record.meta_file[fileidx4input]["frame_idx_start"],
+                record.meta_file[fileidx4input]["frame_idx_end"],
+            )
+            # no target frame to predict for the last frame
+            for frameidx4input in range(
+                endpoints[0],
+                min(endpoints[1], record.num_frames - 1),  # self.num_frames
+            ):
+                input_frame = record.read_frame(
+                    filedesc4input,
+                    record.meta_frame[frameidx4input],
+                    record.read_args["input_features"],
+                )
+                # get the target frame for prediction, both start, stop inclusive
+                lookahead_steps = min(
+                    random.randint(1, record.read_args["max_pred_steps"]),
+                    record.num_frames - 1 - frameidx4input,
+                )
+                frameidx4target = frameidx4input + lookahead_steps
+                target_frame = frame_cache.get(frameidx4target)
+                if target_frame is None:
+                    fileidx4target = record.meta_frame[frameidx4target]["file_idx"]
+                    filedesc4target = file_manager.open_file(
+                        fileidx4target,
+                        file_path=os.path.join(
+                            record.recorddir,
+                            record.meta_file[fileidx4target]["relative_path"],
+                        ),
+                    )
+                    target_frame = record.read_frame(
+                        filedesc4target,
+                        record.meta_frame[frameidx4target],
+                        record.read_args["target_features"],
+                    )
+                # colllate input and target frames so that input and target frame are np.ndarray
+                input_frame = np.vstack(
+                    [
+                        input_frame[feature]
+                        for feature in record.read_args["input_features"]
+                    ]
+                )
+                target_frame = np.vstack(
+                    [
+                        target_frame[feature]
+                        for feature in record.read_args["target_features"]
+                    ]
+                )
+                # print(self.timer.summarize())
+                yield {
+                    "input": input_frame,
+                    "target": target_frame,
+                    "lookahead_steps": np.asarray(lookahead_steps),
+                    "input_features": record.read_args["input_features"],
+                    "target_features": record.read_args["target_features"],
+                }
+        file_manager.close_all_files()
+
+    def dump(self, rank: Optional[int] = None) -> None:
         """save attributes of instance of record into a pickled file and yaml file for visual inspection.
 
         Note:
         saving attribute dict instead of pickled class: pickling class and loading it is a mess because of
         path issues.
         """
+        file_name = f"record_{rank}" if rank is not None else "record_all"
         dic = copy.deepcopy(self.__dict__)
         # do not want to pickle a python module
-        dic["file_desc"] = None
-        with open(os.path.join(self.recorddir, "record.dict"), mode="wb") as f:
+        with open(os.path.join(self.recorddir, f"{file_name}.dict"), mode="wb") as f:
             pickle.dump(dic, file=f)
 
-        # save some attributes of the seqrecord to yaml for human inspection
-        dic["idx_range_of_files"] = None
         # transform some features to make them readable in yaml
-        for key, val in dic["metadata"].items():
+        for _, val in dic["meta_frame"].items():
             for feature in dic["features_written"]:
                 val[feature]["dtype"] = val[feature]["dtype"].str
                 val[feature]["shape"] = list(val[feature]["shape"])
-        with open(os.path.join(self.recorddir, "record_dict.yaml"), mode="w") as f:
+        with open(os.path.join(self.recorddir, f"{file_name}.yaml"), mode="w") as f:
             f.write("# Configs for human inspection only!\n")
             f.write(yaml.dump(dic))
 
     @classmethod
-    def load_record_from_dict(cls, recorddir: str) -> WSR:
+    def load_record_from_dict(cls, recorddir: str, rank: Optional[int] = None) -> WSR:
         """return an instance of sequence record from file that stores attributes of record as a
         dict (stored at path).
 
@@ -613,7 +708,9 @@ class WSeqRecord:
             WSR: an instance of record
         """
 
-        file_path = os.path.join(recorddir, "record.dict")
+        file_path = os.path.join(
+            recorddir, "record_all.dict" if rank is None else f"record_{rank}.dict"
+        )
         with open(file_path, mode="rb") as f:
             obj_dict = pickle.load(f)
         obj = cls(
@@ -624,9 +721,80 @@ class WSeqRecord:
             setattr(obj, key, value)
         return obj
 
+    @classmethod
+    def gather_subseqrecords(
+        cls,
+        recorddir: str,
+        world_size: int,
+        rank2folder: Optional[Dict[int, str]] = None,
+    ) -> WSR:
+        # make everything hierarchical to make it consistent
+        if rank2folder is None:
+            rank2folder = {
+                i: cls.subfolder_name(i, world_size) for i in range(world_size)
+            }
+        sub_records = []
+        for i in range(world_size):
+            sub_records.append(
+                cls.load_record_from_dict(
+                    os.path.join(recorddir, rank2folder[i]), rank=i
+                )
+            )
 
-async def async_open_file(file_idx: int, file_desc_cache: Optional[LRUCache], dir: str):
-    file_desc = await aiofiles.open(recordfileidx2path(dir, file_idx), "rb")
+        # combine meta data
+        features_written = sub_records[0].features_written
+
+        # meta data on each rank collected data
+        meta_rank = {}
+        meta_file = {}
+        meta_frame = {}
+        abs_file_idx = 0
+        abs_frame_idx = 0
+        for i in range(world_size):
+            meta_rank[i] = {
+                "file_idx_start": abs_file_idx,
+                "file_idx_end": abs_file_idx + sub_records[i].num_files,
+                "frame_idx_start": abs_frame_idx,
+            }
+            for j in range(sub_records[i].num_files):
+                meta_file[abs_file_idx] = {
+                    "relative_path": os.path.join(
+                        rank2folder[i],
+                        sub_records[i].meta_file[j]["relative_path"],
+                    ),
+                    "frame_idx_start": abs_frame_idx,
+                }
+                for k in range(
+                    sub_records[i].meta_file[j]["frame_idx_start"],
+                    sub_records[i].meta_file[j]["frame_idx_end"],
+                ):
+                    meta_frame[abs_frame_idx] = sub_records[i].meta_frame[k]
+                    meta_frame[abs_frame_idx]["rel_frame_idx"] = k
+                    meta_frame[abs_frame_idx]["file_idx"] = abs_file_idx
+                    abs_frame_idx += 1
+                meta_file[abs_file_idx]["frame_idx_end"] = abs_frame_idx
+                abs_file_idx += 1
+            meta_rank[i]["frame_idx_end"] = abs_frame_idx
+
+        record = cls(recorddir)
+        record.meta_file = meta_file
+        record.meta_frame = meta_frame
+        record.meta_rank = meta_rank
+        record.features_written = features_written
+
+        record.num_ranks = world_size
+        record.num_files = abs_file_idx
+        record.num_frames = abs_frame_idx
+        return record
+
+    def add_read_args(self, read_args: Dict[str, Any]) -> None:
+        self.read_args = read_args
+
+
+async def async_open_file(
+    file_idx: int, file_desc_cache: Optional[LRUCache], file_path: str
+):
+    file_desc = await aiofiles.open(file_path, "rb")
     if file_desc_cache is not None:
         evicted = file_desc_cache.put(file_idx, file_desc)
         if evicted is not None:
